/*
 *   mod_auth_certificate.c -- client certificate authentication module
 *
 *   Copyright (c) 2009 Johannes Mueller <xjomx@users.sf.net>
 *
 *   This file is part of mod_auth_certificate.
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include "apr_strings.h"
#include "apr_md5.h"            /* for apr_password_validate */
#include "apr_lib.h"            /* for apr_isspace */
#include "apr_base64.h"         /* for apr_base64_decode et al */
#define APR_WANT_STRFUNC        /* for strcasecmp */
#include "apr_want.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_provider.h"

#include "mod_auth_certificate.h"

typedef struct auth_config_struct {
        int enabled;
        int authoritative;
} auth_certificate_config_rec;

static void *create_auth_certificate_dir_config(apr_pool_t *p, char *d)
{
    auth_certificate_config_rec *conf = apr_pcalloc(p, sizeof(*conf));

    conf->enabled = 0;
    /* Any failures are fatal. */
    conf->authoritative = 0;

    return conf;
}

module AP_MODULE_DECLARE_DATA auth_certificate_module;

static int authenticate_user(request_rec * r)
{

	auth_cert_request_rec *notes;
	request_rec *mainreq;

        auth_certificate_config_rec * sec = (auth_certificate_config_rec *)
                ap_get_module_config(r->per_dir_config, &auth_certificate_module);

        /* Ignore request if we are not loaded and enabled */
        if (sec == NULL || sec->enabled != 1) {
		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			"module disabled due to missing CertAuthEnabled directive "
			"or CertAuthEnabled set to off");
                return DECLINED;
        }

	/* always non-authoritative if AuthType Cert is configured */
	if(strcasecmp(ap_auth_type(r), "cert") != 0) {
		sec->authoritative = 0;
	}
	else
	{
		sec->authoritative = 1;
	}

	mainreq = r;
    	while (mainreq->main != NULL) {
        	mainreq = mainreq->main;
    	}
    	while (mainreq->prev != NULL) {
        	mainreq = mainreq->prev;
    	}

	/* make space for the per request config */
    	notes = (auth_cert_request_rec *) ap_get_module_config(mainreq->request_config,
                                                       &auth_certificate_module);
    	if (!notes) {
        	notes = (auth_cert_request_rec *)apr_pcalloc(mainreq->pool, sizeof(auth_cert_request_rec));
        	ap_set_module_config(mainreq->request_config, &auth_certificate_module, notes);
		notes->header_set = 0;
        }

        /* ignore request if not authoritative and authorization header available 
	   and it is our header */

        if (!sec->authoritative && apr_table_get(r->headers_in, "Authorization") && !notes->header_set) {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			"module not needed. Authorization headers already set. "
			"Maybe basic auth in progress.");
                return DECLINED;
        }


        /* Get the clients username from mod_ssl (set SSLUserName) */

        if (r->user == NULL) {
                ap_log_rerror(APLOG_MARK, 
			(sec->authoritative ? APLOG_ERR : APLOG_INFO), 0, r,
			"mod_auth_certificate: No client certificate.");

                if (!sec->authoritative) {

			ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r, 
                      		"mod_auth_certificate: authentication failed, but "
				"non-authoritative. Passing along to other "
				"authentication modules.");

			/* Just returning DECLINED here to use the other modules */
                        return DECLINED;
                }

		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"mod_auth_certificate: authentication failed.");

		/* Returning HTTP_FORBIDDEN because we are authoritiative */
                ap_note_basic_auth_failure(r);
                return HTTP_FORBIDDEN;
        }
        else
        {
		/* mod_ssl set the user correctly */

        	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			"mod_auth_certificate: got username from mod_ssl: '%s'", 
			r->user);
		
		/* set auth type for this request to cert */
		r->ap_auth_type = (char*)apr_pstrdup(r->pool, "cert");

		ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
			"mod_auth_certificate: user %s authenticated successfully. "
			"auth_type: %s", r->user, r->ap_auth_type);

		/* Return OK for every user, because his certificate has been
		 * verified by mod_ssl already */
                return OK;
        }
}

static int set_auth_hdr(request_rec *r)
{
        auth_certificate_config_rec *sec = (auth_certificate_config_rec *)
                ap_get_module_config(r->per_dir_config, &auth_certificate_module);

	const char *auth_line;
	auth_cert_request_rec *notes;
	request_rec *mainreq;

        /* Ignore request if we are not loaded and enabled */
        if (sec == NULL || sec->enabled != 1) {
                return DECLINED;
        }

	mainreq = r;
        while (mainreq->main != NULL) {
                mainreq = mainreq->main;
        }
        while (mainreq->prev != NULL) {
                mainreq = mainreq->prev;
        }

        /* make space for the per request config */
        notes = (auth_cert_request_rec *) ap_get_module_config(mainreq->request_config,
                                                       &auth_certificate_module);
        if (!notes) {
        	notes = (auth_cert_request_rec *)apr_pcalloc(mainreq->pool, sizeof(auth_cert_request_rec));
                ap_set_module_config(mainreq->request_config, &auth_certificate_module, notes);
        }

        /* set a fake authorization header containing username:password if */
        if (strcasecmp(r->ap_auth_type, "cert") == 0 && !apr_table_get(r->headers_in, "Authorization")) {

		if (r->user != NULL) {

			auth_line = apr_pstrcat(r->pool, "Basic ", 
						ap_pbase64encode(r->pool, 
						 	   	 apr_pstrcat(r->pool, r->user,
		                                                 	     ":password", NULL)),
		                                NULL);
		
			apr_table_set(r->headers_in, "Authorization", auth_line);

			ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
                                      "mod_auth_certificate: r->ap_auth_type is cert "
				      "and user is available. Setting fake authorization "
				      "header: \"Authorization: %s\"",
   			              auth_line);

			/* Note that header was set by us */
			notes->header_set = 1;
		}
	}

        /* We don't really check access here */
        return DECLINED;
}

int initModule(apr_pool_t *p, apr_pool_t *plog,
                    apr_pool_t *ptemp,
                    server_rec *s){

	void *data;
    	const char *userdata_key = "mod_auth_certificate_init";

	/* initModule() will be called twice. Don't bother
	 * going through all of the initialization on the first call
	 * because it will just be thrown away.*/
	apr_pool_userdata_get(&data, userdata_key, s->process->pool);
	if (!data) {
	    apr_pool_userdata_set((const void *)1, userdata_key,
	                           apr_pool_cleanup_null, s->process->pool);
	
	    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
	                	 "mod_auth_certificate/0.3 by Johannes Mueller");
	}

	return OK;
}

static const command_rec auth_certificate_cmds[] = {
        AP_INIT_FLAG("CertAuthEnabled",
                        ap_set_flag_slot,
                        (void *) APR_OFFSETOF(auth_certificate_config_rec, enabled),
                        OR_AUTHCFG,
                        "Set to 'on' to enable client certificate authentication"),
        {NULL}
};

module AP_MODULE_DECLARE_DATA auth_certificate_module;


static void register_hooks(apr_pool_t *p)
{
	static const char *aszPre[] = { "mod_ssl.c", NULL };
	ap_hook_post_config(initModule, NULL,NULL, APR_HOOK_MIDDLE);
        ap_hook_check_user_id(authenticate_user, aszPre, NULL, APR_HOOK_FIRST);
	ap_hook_auth_checker(set_auth_hdr, aszPre, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA auth_certificate_module =
{
    STANDARD20_MODULE_STUFF,
    create_auth_certificate_dir_config,  /* dir config creater */
    NULL,                         	 /* dir merger --- default is to override */
    NULL,                         	 /* server config */
    NULL,                         	 /* merge server config */
    auth_certificate_cmds,               /* command apr_table_t */
    register_hooks                 	/* register hooks */
};
